# ------- #
# Imports #
# ------- #

from .fns import discardLastWhile, passThrough
import re


# ---- #
# Init #
# ---- #

onlyWhitespaceRe = re.compile(r"\s*$")


# ---- #
# Main #
# ---- #


def areOnlyWhitespace(aString):
    return isOnlyWhitespace(aString)


def getNumberOfLeadingSpaces(line):
    i = 0

    for c in line:
        if c != " ":
            break
        i += 1

    return i


def isOnlyWhitespace(aString):
    return onlyWhitespaceRe.match(aString) is not None


def removeFirstAndLast(aList):
    aList.pop(0)
    aList.pop()
    return aList


#
# - removes any tailing lines which only contain whitespace
# - last line with a non-whitespace character has tailing whitespace removed
#
def trimLastLines(lines):
    return passThrough(
        lines, [discardLastWhile(isOnlyWhitespace), trimTheLastLine]
    )


# ------- #
# Helpers #
# ------- #


def trimTheLastLine(allLines):
    allLines[-1] = allLines[-1].rstrip()
    return allLines



# ------- #
# Imports #
# ------- #

from .fns import all_, isEmpty, joinWith, passThrough, raise_
from .indentWith import indentWith
import os

from .utils import (
    areOnlyWhitespace,
    getNumberOfLeadingSpaces,
    isOnlyWhitespace,
    removeFirstAndLast,
    trimLastLines,
)

from .invalid import (
    invalidFewerThan3Lines,
    invalidFirstOrLastLine,
    invalidSecondLine,
)

# ---- #
# Main #
# ---- #


def tedent(aString):
    validateInput(aString)

    if isEmpty(aString):
        return ""

    allLines = aString.split(os.linesep)

    if len(allLines) < 3:
        if all_(areOnlyWhitespace)(allLines):
            return ""
        else:
            raise ValueError(invalidFewerThan3Lines)

    firstLine = allLines[0]
    secondLine = allLines[1]
    lastLine = allLines[-1]

    if not all_(areOnlyWhitespace)([firstLine, lastLine]):
        raise ValueError(invalidFirstOrLastLine)
    elif isOnlyWhitespace(secondLine):
        raise ValueError(invalidSecondLine)

    #
    # we got valid input fam
    #

    anchor = getNumberOfLeadingSpaces(secondLine)

    # allLines is being mutated for performance
    return passThrough(
        allLines,
        [removeFirstAndLast, indentWith(anchor), trimLastLines, joinWith("\n")],
    )


# ------- #
# Helpers #
# ------- #


def validateInput(input):
    if not isinstance(input, str):
        raise_(
            TypeError,
            f"""
            tedent requires a string

            type given: {type(input).__name__}
            """,
        )



from textwrap import dedent

expectedUsage = """
    Expected usage:
    tedent(f\"\"\"
        Some string
        {here}
    \"\"\")
    """

invalidFewerThan3Lines = dedent(
    f"""
    tedent expects a string with three or more lines.  When fewer are passed
    then they must contain only whitespace for this error not to be thrown.
    {expectedUsage}
    """
)

invalidFirstOrLastLine = dedent(
    f"""
    The first and last lines are only allowed to contain whitespace
    {expectedUsage}
    """
)

invalidSecondLine = dedent(
    f"""
    The second line must have a non-whitespace character
    {expectedUsage}
    """
)



from from ..._vendor.ordered_set import OrderedSet
from ..._vendor import wrapt

from ..internal.discardWhen import discardWhen
from ..internal.get import get
from ..internal.getArgName import getArgName
from ..internal.isLaden import isLaden
from ..internal.isType import isType
from ..internal.joinWith import joinWith
from ..internal.map_ import map_
from ..internal.passThrough import passThrough
from ..internal.raise_ import raise_
from ..internal.sort import sort
from ..internal.toType import toType


def argIsListOfType(aType):
    @wrapt.decorator
    def wrapper(fn, _instance, args, kwargs):
        typePassed = type(args[0])

        fnName = fn.__name__
        typeName = aType.__name__

        if typePassed is not list:
            argName = getArgName(fn)
            raise_(
                ValueError,
                f"""\
                {fnName} requires {argName} to have the type list
                type passed: {typePassed.__name__}
                """,
            )

        invalidTypes = discardWhen(isType(aType))(args[0])
        if isLaden(invalidTypes):
            argName = getArgName(fn)
            invalidTypeNames = passThrough(
                invalidTypes,
                [
                    map_(toType),
                    OrderedSet,
                    list,
                    map_(get("__name__")),
                    sort,
                    joinWith(", "),
                ],
            )
            raise_(
                ValueError,
                f"""\
                {fnName} requires {argName} to be a list of {typeName}
                invalid types passed: {invalidTypeNames}
                """,
            )

        return fn(*args, **kwargs)

    return wrapper



from .all_ import all_
from .discardFirst import discardFirst
from .discardLastWhile import discardLastWhile
from .forEach import forEach
from .iif import iif
from .invokeAttr import invokeAttr
from .isEmpty import isEmpty
from .joinWith import joinWith
from .passThrough import passThrough
from .raise_ import raise_

__all__ = [
    "all_",
    "discardFirst",
    "discardLastWhile",
    "forEach",
    "iif",
    "invokeAttr",
    "isEmpty",
    "joinWith",
    "passThrough",
    "raise_",
]



# ------- #
# Imports #
# ------- #

from .decorators.argIsCallable import argIsCallable
from .internal.getTypedResult import getTypedResult
from .internal.makeGenericCallFn import makeGenericCallFn


# ---- #
# Main #
# ---- #


@argIsCallable
def discardLastWhile(predicate):
    fnName = discardLastWhile.__name__
    shouldDiscard = makeGenericCallFn(predicate, 3, fnName)

    def discardLastWhile_inner(collection):
        discardLastWhileFn = getTypedResult(
            collection, typeToDiscardLastWhile, fnName
        )
        return discardLastWhileFn(shouldDiscard, collection)

    return discardLastWhile_inner


# ------- #
# Helpers #
# ------- #


def discardLastWhile_viaSlice(shouldDiscard, sliceAble):
    i = len(sliceAble)
    for el in reversed(sliceAble):
        if not shouldDiscard(el, i, sliceAble):
            break

        i -= 1

    return sliceAble[:i]


typeToDiscardLastWhile = {
    list: discardLastWhile_viaSlice,
    str: discardLastWhile_viaSlice,
}



# ------- #
# Imports #
# ------- #

from .fns import discardFirst, passThrough
from .utils import getNumberOfLeadingSpaces


# ---- #
# Main #
# ---- #


def indentWith(anchor):
    def indentWith_inner(allLines):
        if anchor == 0:
            return allLines

        runningIndent = 0

        for i, oldLine in enumerate(allLines):
            allLines[i] = adjustWhitespace(allLines[i], anchor, runningIndent)
            runningIndent = passThrough(
                oldLine, [getNumberOfLeadingSpaces, updateIndent(runningIndent)]
            )

        return allLines

    def updateIndent(runningIndent):
        def updateIndent_inner(leadingSpaces):
            maybeNewIndent = leadingSpaces - anchor

            if maybeNewIndent == 0:
                return 0
            elif maybeNewIndent > 0:
                return maybeNewIndent
            else:
                return runningIndent

        return updateIndent_inner

    return indentWith_inner


# ------- #
# Helpers #
# ------- #


def adjustWhitespace(line, anchor, runningIndent):
    numberOfLeadingSpaces = getNumberOfLeadingSpaces(line)
    lineWithoutLeadingSpace = discardFirst(numberOfLeadingSpaces)(line)
    currentIndent = numberOfLeadingSpaces - anchor

    newLeadingSpace = 0

    if currentIndent > 0:
        newLeadingSpace = currentIndent
    elif currentIndent < 0:
        newLeadingSpace = numberOfLeadingSpaces + runningIndent

    return (" " * newLeadingSpace) + lineWithoutLeadingSpace



# ------- #
# Imports #
# ------- #

from .internal.getTypedResult import getTypedResult


# ---- #
# Main #
# ---- #


def discardFirst(n):
    fnName = discardFirst.__name__

    def discardFirst_inner(collection):
        discardFn = getTypedResult(collection, typeToDiscardFirst, fnName)
        return discardFn(n, collection)

    return discardFirst_inner


# ------- #
# Helpers #
# ------- #


def discardFirst_viaSlice(n, sliceAble):
    return sliceAble[n:]


typeToDiscardFirst = {list: discardFirst_viaSlice, str: discardFirst_viaSlice}



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['tedent',
 'tedent._vendor',
 'tedent._vendor.wrapt',
 'tedent.fns',
 'tedent.fns.decorators',
 'tedent.fns.internal']

package_data = \
{'': ['*'],
 'tedent._vendor': ['ordered_set-3.1.1-py3.7.egg-info/*',
                    'wrapt-1.11.1.dist-info/*']}

setup_kwargs = {
    'name': 'tedent',
    'version': '0.1.3',
    'description': 'like dedent but more flexible',
    'long_description': '# Tedent\n\nKeep your multi-line templated strings lookin\' good :sunglasses:\n\n*This is a python version of [tedent](https://github.com/olsonpm/tedent)*\n\n<br>\n\n<!-- START doctoc generated TOC please keep comment here to allow auto update -->\n<!-- DON\'T EDIT THIS SECTION, INSTEAD RE-RUN doctoc TO UPDATE -->\n**Table of Contents**\n\n- [What is it?](#what-is-it)\n- [What does the name stand for?](#what-does-the-name-stand-for)\n- [Why create it?](#why-create-it)\n- [Simple Usage](#simple-usage)\n- [Questions about how the indentation works?](#questions-about-how-the-indentation-works)\n- [Important Usage Notes](#important-usage-notes)\n  - [input requirements](#edge-cases-and-input-requirements)\n- [Test](#test)\n\n<!-- END doctoc generated TOC please keep comment here to allow auto update -->\n\n<br>\n\n### What is it?\n\n- A function similar to [dedent](https://docs.python.org/3.7/library/textwrap.html#textwrap.dedent)\n  just with different semantics\n\n<br>\n\n### What does the name stand for?\n\n- `Te`mplate string\n- in`dent`ation\n\nnames are hard\n\n<br>\n\n### Why create it?\n\n- dedent didn\'t handle the following case like I wanted\n\n```py\nformattedBoroughs = f"""\\\n[\n    \'Brooklyn\',\n    \'Manhattan\',\n]\n"""\n\nprint(\n    dedent(\n        f"""\\\n        New York boroughs\n        ${formattedBoroughs}\n        """\n    )\n)\n\n#\n# expected:\n# New York boroughs\n# [\n#     \'Brooklyn\',\n#     \'Manhattan\',\n# ]\n#\n# actual:\n#         New York boroughs\n#         [\n#     \'Brooklyn\',\n#     \'Manhattan\',\n# ]\n#\n```\n\n<br>\n\n### Simple Usage\n\n```py\nimport tedent from \'tedent\'\n\n#\n# *note the lack of the backslash\n#\nprint(\n    tedent(\n        """\n        This will be indented\n          as you expect\n        """\n    )\n)\n\n# writes:\n# This will be indented\n#   as you expect\n```\n\n<br>\n\n### Questions about how the indentation works?\n\nBecause the indentation logic is both young and convoluted, please refer to\n[the code](tedent/index.py) and [tests](tests/simple.py) for details. The\nlibrary is not that big and if you have any questions please create a\ngithub issue.\n\n<br>\n\n### Important Usage Notes\n\n- First of all, this library doesn\'t handle tabs. I will accept a PR\n  with support\n\n- Secondly, if you always use `tedent` like the following\n\n  ```py\n  tedent(\n      """\n      some text\n      """\n  )\n  ```\n\n  then you shouldn\'t run into any issues. However we all know input can be\n  tricky so `tedent` has a few input requirements in order to format your\n  string properly.\n\n#### input requirements\n\n- if the argument isn\'t a string then an error will be thrown\n- if you pass a string with three or more newlines, then\n  - the first and last lines must contain only whitespace\n  - the second line must contain a non-whitespace character\n  - _an error will be thrown if the above two conditions are not met_\n- if you pass a string with fewer than 3 newlines\n  - if they only contain whitespace then an empty string is returned\n  - otherwise an error is thrown\n- finally, all trailing whitespace from the result is stripped\n\nIf you have questions please raise a github issue.\n\n<br>\n\n### Test\n\n```py\n#\n# you must have poetry installed\n#\n$ poetry shell\n$ poetry install\n$ python runTests.py\n```\n',
    'author': 'Philip Olson',
    'author_email': 'philip.olson@pm.me',
    'url': 'https://github.com/olsonpm/py_tedent',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
